@@ -27,6 +27,7 @@ struct command {
 	const char	*format;
 };
 
+static int act_keyctl___version(int argc, char *argv[]);
 static int act_keyctl_show(int argc, char *argv[]);
 static int act_keyctl_add(int argc, char *argv[]);
 static int act_keyctl_padd(int argc, char *argv[]);
@@ -63,6 +64,7 @@ static int act_keyctl_reap(int argc, char *argv[]);
 static int act_keyctl_purge(int argc, char *argv[]);
 
 const struct command commands[] = {
+	{ act_keyctl___version,	"--version",	"" },
 	{ act_keyctl_add,	"add",		"<type> <desc> <data> <keyring>" },
 	{ act_keyctl_chgrp,	"chgrp",	"<key> <gid>" },
 	{ act_keyctl_chown,	"chown",	"<key> <uid>" },
@@ -97,14 +99,14 @@ const struct command commands[] = {
 	{ NULL,			"session",	"- [<prog> <arg1> <arg2> ...]" },
 	{ NULL,			"session",	"<name> [<prog> <arg1> <arg2> ...]" },
 	{ act_keyctl_setperm,	"setperm",	"<key> <mask>" },
-	{ act_keyctl_show,	"show",		"" },
+	{ act_keyctl_show,	"show",		"[-x] [<keyring>]" },
 	{ act_keyctl_timeout,	"timeout",	"<key> <timeout>" },
 	{ act_keyctl_unlink,	"unlink",	"<key> [<keyring>]" },
 	{ act_keyctl_update,	"update",	"<key> <data>" },
 	{ NULL,			NULL,		NULL }
 };
 
-static int dump_key_tree(key_serial_t keyring, const char *name);
+static int dump_key_tree(key_serial_t keyring, const char *name, int hex_key_IDs);
 static void format(void) __attribute__((noreturn));
 static void error(const char *msg) __attribute__((noreturn));
 static key_serial_t get_key_id(const char *arg);
@@ -221,11 +223,22 @@ static void format(void)
 
 } /* end format() */
 
+/*****************************************************************************/
+/*
+ * Display version information
+ */
+static int act_keyctl___version(int argc, char *argv[])
+{
+	printf("keyctl from %s (Built %s)\n",
+	       keyutils_version_string, keyutils_build_string);
+	return 0;
+}
+
 /*****************************************************************************/
 /*
  * grab data from stdin
  */
-static char *grab_stdin(void)
+static char *grab_stdin(size_t *_size)
 {
 	static char input[65536 + 1];
 	int n, tmp;
@@ -249,6 +262,7 @@ static char *grab_stdin(void)
 	}
 
 	input[n] = '\0';
+	*_size = n;
 
 	return input;
 
@@ -306,10 +320,22 @@ write_mask:
  */
 static int act_keyctl_show(int argc, char *argv[])
 {
-	if (argc != 1)
+	key_serial_t keyring = KEY_SPEC_SESSION_KEYRING;
+	int hex_key_IDs = 0;
+
+	if (argc >= 2 && strcmp(argv[1], "-x") == 0) {
+		hex_key_IDs = 1;
+		argc--;
+		argv++;
+	}
+
+	if (argc > 2)
 		format();
 
-	dump_key_tree(KEY_SPEC_SESSION_KEYRING, "Session Keyring");
+	if (argc == 2)
+		keyring = get_key_id(argv[1]);
+
+	dump_key_tree(keyring, argc == 2 ? "Keyring" : "Session Keyring", hex_key_IDs);
 	return 0;
 
 } /* end act_keyctl_show() */
@@ -344,19 +370,26 @@ static int act_keyctl_add(int argc, char *argv[])
  */
 static int act_keyctl_padd(int argc, char *argv[])
 {
-	char *args[6];
+	key_serial_t dest;
+	size_t datalen;
+	void *data;
+	int ret;
+
 
 	if (argc != 4)
 		format();
 
-	args[0] = argv[0];
-	args[1] = argv[1];
-	args[2] = argv[2];
-	args[3] = grab_stdin();
-	args[4] = argv[3];
-	args[5] = NULL;
+	dest = get_key_id(argv[3]);
 
-	return act_keyctl_add(5, args);
+	data = grab_stdin(&datalen);
+
+	ret = add_key(argv[1], argv[2], data, datalen, dest);
+	if (ret < 0)
+		error("add_key");
+
+	/* print the resulting key ID */
+	printf("%d\n", ret);
+	return 0;
 
 } /* end act_keyctl_padd() */
 
@@ -420,6 +453,7 @@ static int act_keyctl_request2(int argc, char *argv[])
 static int act_keyctl_prequest2(int argc, char *argv[])
 {
 	char *args[6];
+	size_t datalen;
 
 	if (argc != 3 && argc != 4)
 		format();
@@ -427,7 +461,7 @@ static int act_keyctl_prequest2(int argc, char *argv[])
 	args[0] = argv[0];
 	args[1] = argv[1];
 	args[2] = argv[2];
-	args[3] = grab_stdin();
+	args[3] = grab_stdin(&datalen);
 	args[4] = argv[3];
 	args[5] = NULL;
 
@@ -461,17 +495,20 @@ static int act_keyctl_update(int argc, char *argv[])
  */
 static int act_keyctl_pupdate(int argc, char *argv[])
 {
-	char *args[4];
+	key_serial_t key;
+	size_t datalen;
+	void *data;
 
 	if (argc != 2)
 		format();
 
-	args[0] = argv[0];
-	args[1] = argv[1];
-	args[2] = grab_stdin();
-	args[3] = NULL;
+	key = get_key_id(argv[1]);
+	data = grab_stdin(&datalen);
 
-	return act_keyctl_update(3, args);
+	if (keyctl_update(key, data, datalen) < 0)
+		error("keyctl_update");
+
+	return 0;
 
 } /* end act_keyctl_pupdate() */
 
@@ -1134,18 +1171,21 @@ static int act_keyctl_instantiate(int argc, char *argv[])
  */
 static int act_keyctl_pinstantiate(int argc, char *argv[])
 {
-	char *args[5];
+	key_serial_t key, dest;
+	size_t datalen;
+	void *data;
 
 	if (argc != 3)
 		format();
 
-	args[0] = argv[0];
-	args[1] = argv[1];
-	args[2] = grab_stdin();
-	args[3] = argv[2];
-	args[4] = NULL;
+	key = get_key_id(argv[1]);
+	dest = get_key_id(argv[2]);
+	data = grab_stdin(&datalen);
 
-	return act_keyctl_instantiate(4, args);
+	if (keyctl_instantiate(key, data, datalen, dest) < 0)
+		error("keyctl_instantiate");
+
+	return 0;
 
 } /* end act_keyctl_pinstantiate() */
 
@@ -1550,7 +1590,7 @@ static key_serial_t get_key_id(const char *arg)
 /*
  * recursively display a key/keyring tree
  */
-static int dump_key_tree_aux(key_serial_t key, int depth, int more)
+static int dump_key_tree_aux(key_serial_t key, int depth, int more, int hex_key_IDs)
 {
 	static char dumpindent[64];
 	key_serial_t *pk;
@@ -1604,13 +1644,22 @@ static int dump_key_tree_aux(key_serial_t key, int depth, int more)
 	/* and print */
 	calc_perms(pretty_mask, perm, uid, gid);
 
-	printf("%9d %s  %5d %5d  %s%s%s: %s\n",
-	       key,
-	       pretty_mask,
-	       uid, gid,
-	       dumpindent,
-	       depth > 0 ? "\\_ " : "",
-	       type, desc + dpos);
+	if (hex_key_IDs)
+		printf("0x%08x %s  %5d %5d  %s%s%s: %s\n",
+		       key,
+		       pretty_mask,
+		       uid, gid,
+		       dumpindent,
+		       depth > 0 ? "\\_ " : "",
+		       type, desc + dpos);
+	else
+		printf("%10d %s  %5d %5d  %s%s%s: %s\n",
+		       key,
+		       pretty_mask,
+		       uid, gid,
+		       dumpindent,
+		       depth > 0 ? "\\_ " : "",
+		       type, desc + dpos);
 
 	/* if it's a keyring then we're going to want to recursively
 	 * display it if we can */
@@ -1661,7 +1710,8 @@ static int dump_key_tree_aux(key_serial_t key, int depth, int more)
 
 				kcount += dump_key_tree_aux(key,
 							    rdepth,
-							    ringlen - 4 >= sizeof(key_serial_t));
+							    ringlen - 4 >= sizeof(key_serial_t),
+							    hex_key_IDs);
 			}
 
 		} while (ringlen -= 4, ringlen >= sizeof(key_serial_t));
@@ -1678,9 +1728,14 @@ static int dump_key_tree_aux(key_serial_t key, int depth, int more)
 /*
  * recursively list a keyring's contents
  */
-static int dump_key_tree(key_serial_t keyring, const char *name)
+static int dump_key_tree(key_serial_t keyring, const char *name, int hex_key_IDs)
 {
 	printf("%s\n", name);
-	return dump_key_tree_aux(keyring, 0, 0);
+
+	keyring = keyctl_get_keyring_ID(keyring, 0);
+	if (keyring == -1)
+		error("Unable to dump key");
+
+	return dump_key_tree_aux(keyring, 0, 0, hex_key_IDs);
 
 } /* end dump_key_tree() */
